import capstone as _capstone

try:
    import unicorn as _unicorn
except ImportError:
    _unicorn = None

from .arch import Arch
from .archerror import ArchError

class ArchX86(Arch):
    def __init__(self, endness='Iend_LE'):
        if endness != 'Iend_LE':
            raise ArchError('Arch i386 must be little endian')
        super(ArchX86, self).__init__(endness)
        if self.vex_archinfo:
            self.vex_archinfo['x86_cr0'] = 0xFFFFFFFF

    bits = 32
    vex_arch = "VexArchX86"
    name = "X86"
    qemu_name = 'i386'
    ida_processor = 'metapc'
    linux_name = 'i386'
    triplet = 'i386-linux-gnu'
    max_inst_bytes = 15
    call_sp_fix = -8
    ip_offset = 68
    sp_offset = 24
    bp_offset = 28
    ret_offset = 8
    syscall_num_offset = 8
    call_pushes_ret = True
    stack_change = -4
    memory_endness = "Iend_LE"
    register_endness = "Iend_LE"
    sizeof = {'int': 32, 'long': 32, 'long long': 64}
    cs_arch = _capstone.CS_ARCH_X86
    cs_mode = _capstone.CS_MODE_32 + _capstone.CS_MODE_LITTLE_ENDIAN
    uc_arch = _unicorn.UC_ARCH_X86 if _unicorn else None
    uc_mode = (_unicorn.UC_MODE_32 + _unicorn.UC_MODE_LITTLE_ENDIAN) if _unicorn else None
    uc_const = _unicorn.x86_const if _unicorn else None
    uc_prefix = "UC_X86_" if _unicorn else None
    function_prologs = {
        r"\x55\x8b\xec", # push ebp; mov ebp, esp
        r"\x55\x89\xe5",  # push ebp; mov ebp, esp
    }
    function_epilogs = {
        r"\xc9\xc3", # leave; ret
        r"([^\x41][\x50-\x5f]{1}|\x41[\x50-\x5f])\xc3", # pop <reg>; ret
        r"[^\x48][\x83,\x81]\xc4([\x00-\xff]{1}|[\x00-\xff]{4})\xc3", #  add esp, <siz>; retq
    }
    ret_instruction = "\xc3"
    nop_instruction = "\x90"
    instruction_alignment = 1
    default_register_values = [
        ( 'esp', Arch.initial_sp, True, 'global' ), # the stack
        ( 'd', 1, False, None ),
        ( 'fpround', 0, False, None ),
        ( 'sseround', 0, False, None ),
        ( 'gdt', 0, False, None ),
        ( 'ldt', 0, False, None ),
        ( 'id', 1, False, None ),
        ( 'ac', 0, False, None ),
        ( 'ftop', 0, False, None ),
        ( 'fpu_tags', 0, False, None)
    ]
    entry_register_values = {
        'eax': 0x1C,
        'edx': 'ld_destructor',
        'ebp': 0
    }
    default_symbolic_registers = [ 'eax', 'ecx', 'edx', 'ebx', 'esp', 'ebp', 'esi', 'edi', 'eip' ]
    register_names = {
        8: 'eax',
        12: 'ecx',
        16: 'edx',
        20: 'ebx',

        24: 'esp',

        28: 'ebp',
        32: 'esi',
        36: 'edi',

        # condition stuff
        40: 'cc_op',
        44: 'cc_dep1',
        48: 'cc_dep2',
        52: 'cc_ndep',

        # this determines which direction SSE instructions go
        56: 'd',

        # separately-stored bits of eflags
        60: 'id',
        64: 'ac',

        68: 'eip',

        # fpu registers
        72: 'st7',
        80: 'st6',
        88: 'st5',
        96: 'st4',
        104: 'st3',
        112: 'st2',
        120: 'st1',
        128: 'st0',

        # fpu tags
        136: 'fpu_t0',
        137: 'fpu_t1',
        138: 'fpu_t2',
        139: 'fpu_t3',
        140: 'fpu_t4',
        141: 'fpu_t5',
        142: 'fpu_t6',
        143: 'fpu_t7',

        # fpu settings
        144: 'fpround',
        148: 'fc3210',
        152: 'ftop',

        # sse
        156: 'sseround',
        160: 'xmm0',
        176: 'xmm1',
        192: 'xmm2',
        208: 'xmm3',
        224: 'xmm4',
        240: 'xmm5',
        256: 'xmm6',
        272: 'xmm7',

        288: 'cs',
        290: 'ds',
        292: 'es',
        294: 'fs',
        296: 'gs',
        298: 'ss',

        304: 'ldt',
        312: 'gdt'
    }

    registers = {
        'eax': (8, 4),
        'ecx': (12, 4),
        'edx': (16, 4),
        'ebx': (20, 4),

        'sp': (24, 4),
        'esp': (24, 4),

        'ebp': (28, 4), 'bp': (28, 4),
        'esi': (32, 4),
        'edi': (36, 4),

        # condition stuff
        'cc_op': (40, 4),
        'cc_dep1': (44, 4),
        'cc_dep2': (48, 4),
        'cc_ndep': (52, 4),

        # this determines which direction SSE instructions go
        'd': (56, 4),

        # separately-stored bits of eflags
        'id': (60, 4),
        'ac': (64, 4),

        'eip': (68, 4),
        'pc': (68, 4),
        'ip': (68, 4),

        # fpu registers and mmx aliases
        'fpu_regs': (72, 64),
        'st7': (72, 8),
        'st6': (80, 8),
        'st5': (88, 8),
        'st4': (96, 8),
        'st3': (104, 8),
        'st2': (112, 8),
        'st1': (120, 8),
        'st0': (128, 8),
        'mm0': (72, 8),
        'mm1': (80, 8),
        'mm2': (88, 8),
        'mm3': (96, 8),
        'mm4': (104, 8),
        'mm5': (112, 8),
        'mm6': (120, 8),
        'mm7': (128, 8),

        # fpu tags
        'fpu_tags': (136, 8),
        'fpu_t0': (136, 1),
        'fpu_t1': (137, 1),
        'fpu_t2': (138, 1),
        'fpu_t3': (139, 1),
        'fpu_t4': (140, 1),
        'fpu_t5': (141, 1),
        'fpu_t6': (142, 1),
        'fpu_t7': (143, 1),

        # fpu settings
        'fpround': (144, 4),
        'fc3210': (148, 4),
        'ftop': (152, 4),

        # sse
        'sseround': (156, 4),
        'xmm0': (160, 16),
        'xmm1': (176, 16),
        'xmm2': (192, 16),
        'xmm3': (208, 16),
        'xmm4': (224, 16),
        'xmm5': (240, 16),
        'xmm6': (256, 16),
        'xmm7': (272, 16),

        'cs': (288, 2),
        'ds': (290, 2),
        'es': (292, 2),
        'fs': (294, 2),
        'gs': (296, 2),
        'ss': (298, 2),
        'ldt': (304, 8),
        'gdt': (312, 8)
    }

    argument_registers = { registers['eax'][0],
                           registers['ecx'][0],
                           registers['edx'][0],
                           registers['ebx'][0],
                           registers['ebp'][0],
                           registers['esi'][0],
                           registers['edi'][0] }

    lib_paths = ['/lib32', '/usr/lib32']
    got_section_name = '.got.plt'
    ld_linux_name = 'ld-linux.so.2'



import capstone as _capstone

#try:
#    import unicorn as _unicorn
#except ImportError:
#    _unicorn = None

from .arch import Arch

# Note: PowerPC doesn't have pc, so guest_CIA is commented as IP (no arch visible register)
# Normally r1 is used as stack pointer

class ArchPPC64(Arch):
    def __init__(self, endness="Iend_LE"):
        super(ArchPPC64, self).__init__(endness)
        if endness == 'Iend_BE':
            self.function_prologs = {
                r"\x94\x21[\x00-\xff]{2}\x7c\x08\x02\xa6",                        # stwu r1, -off(r1); mflr r0
            }
            self.function_epilogs = {
                r"[\x00-\xff]{2}\x03\xa6([\x00-\xff]{4}){0,6}\x4e\x80\x00\x20"    # mtlr reg; ... ; blr
            }
            self.triplet = 'powerpc-linux-gnu'

    bits = 64
    vex_arch = "VexArchPPC64"
    name = "PPC64"
    qemu_name = 'ppc64'
    ida_processor = 'ppc64'
    triplet = 'powerpc64le-linux-gnu'
    linux_name = 'ppc750'
    max_inst_bytes = 4
    ip_offset = 1296
    sp_offset = 24
    bp_offset = 264
    ret_offset = 40
    syscall_num_offset = 16
    call_pushes_ret = False
    stack_change = -8
    initial_sp = 0xffffffffff000000
    sizeof = {'int': 32, 'long': 64, 'long long': 64}
    cs_arch = _capstone.CS_ARCH_PPC
    cs_mode = _capstone.CS_MODE_64 + _capstone.CS_MODE_LITTLE_ENDIAN
    # unicorn not supported
    #uc_arch = _unicorn.UC_ARCH_PPC if _unicorn else None
    #uc_mode = (_unicorn.UC_MODE_64 + _unicorn.UC_MODE_LITTLE_ENDIAN) if _unicorn else None
    ret_instruction = "\x20\x00\x80\x4e"
    nop_instruction = "\x00\x00\x00\x60"
    instruction_alignment = 4
    persistent_regs = [ 'r2' ]

    function_prologs = {
        r"[\x00-\xff]{2}\x21\x94\xa6\x02\x08\x7c",                        # stwu r1, -off(r1); mflr r0
    }
    function_epilogs = {
        r"\xa6\x03[\x00-\xff]{2}([\x00-\xff]{4}){0,6}\x20\x00\x80\x4e"    # mtlr reg; ... ; blr
    }

    default_register_values = [
        ( 'sp', initial_sp, True, 'global' ) # the stack
    ]
    entry_register_values = {
        'r2': 'toc',
        'r3': 'argc',
        'r4': 'argv',
        'r5': 'envp',
        'r6': 'auxv',
        'r7': 'ld_destructor'
    }

    register_names = {
        16: 'r0',
        24: 'r1',
        32: 'r2',
        40: 'r3',
        48: 'r4',
        56: 'r5',
        64: 'r6',
        72: 'r7',
        80: 'r8',
        88: 'r9',
        96: 'r10',
        104: 'r11',
        112: 'r12',
        120: 'r13',
        128: 'r14',
        136: 'r15',
        144: 'r16',
        152: 'r17',
        160: 'r18',
        168: 'r19',
        176: 'r20',
        184: 'r21',
        192: 'r22',
        200: 'r23',
        208: 'r24',
        216: 'r25',
        224: 'r26',
        232: 'r27',
        240: 'r28',
        248: 'r29',
        256: 'r30',
        264: 'r31',

        1296: 'pc',
        1302: 'lr'
    }

    registers = {
        'r0': (16, 8),
        'r1': (24, 8), 'sp': (24, 8),
        'r2': (32, 8), 'rtoc': (32, 8),
        'r3': (40, 8),
        'r4': (48, 8),
        'r5': (56, 8),
        'r6': (64, 8),
        'r7': (72, 8),
        'r8': (80, 8),
        'r9': (88, 8),
        'r10': (96, 8),
        'r11': (104, 8),
        'r12': (112, 8),
        'r13': (120, 8),
        'r14': (128, 8),
        'r15': (136, 8), 'bp': (136, 8),
        'r16': (144, 8),
        'r17': (152, 8),
        'r18': (160, 8),
        'r19': (168, 8),
        'r20': (176, 8),
        'r21': (184, 8),
        'r22': (192, 8),
        'r23': (200, 8),
        'r24': (208, 8),
        'r25': (216, 8),
        'r26': (224, 8),
        'r27': (232, 8),
        'r28': (240, 8),
        'r29': (248, 8),
        'r30': (256, 8),
        'r31': (260, 8),

        'ip': (1296, 8), 'pc': (1296, 8),
        'lr': (1304, 8)
    }

    argument_registers = {
        registers['r0'],
        registers['r2'],
        registers['r3'],
        registers['r4'],
        registers['r5'],
        registers['r6'],
        registers['r7'],
        registers['r8'],
        registers['r9'],
        registers['r10'],
        registers['r11'],
        registers['r12'],
        registers['r13'],
        registers['r14'],
        registers['r15'],
        registers['r16'],
        registers['r17'],
        registers['r18'],
        registers['r19'],
        registers['r20'],
        registers['r21'],
        registers['r22'],
        registers['r23'],
        registers['r24'],
        registers['r25'],
        registers['r26'],
        registers['r27'],
        registers['r28'],
        registers['r29'],
        registers['r30'],
        registers['r31'],
    }

    got_section_name = '.plt'
    ld_linux_name = 'ld64.so.1'



"""
archinfo is a collection of classes that contain architecture-specific information.
It is useful for cross-architecture tools (such as pyvex).
"""
# pylint: disable=wildcard-import
from .arch import *
from .defines import defines



import capstone as _capstone
import struct as _struct

try:
    import pyvex as _pyvex
except ImportError:
    _pyvex = None

try:
    import unicorn as _unicorn
except ImportError:
    _unicorn = None

import logging
l = logging.getLogger('archinfo.arch')
l.addHandler(logging.NullHandler())

class Arch(object):
    """
    Represent an architecture. This class is responsible for architecture-specific things such as call emulation.

    :ivar name: The name of the architecture.
    :type name: str

    """
    def __init__(self, endness):
        if endness not in ('Iend_LE', 'Iend_BE'):
            raise ArchError('Must pass a valid VEX endness: "Iend_LE" or "Iend_BE"')

        if _pyvex:
            self.vex_archinfo = _pyvex.default_vex_archinfo()
        if endness == 'Iend_BE':
            if self.vex_archinfo:
                self.vex_archinfo['endness'] = _pyvex.vex_endness_from_string('VexEndnessBE')
            self.memory_endness = 'Iend_BE'
            self.register_endness = 'Iend_BE'
            self.cs_mode -= _capstone.CS_MODE_LITTLE_ENDIAN
            self.cs_mode += _capstone.CS_MODE_BIG_ENDIAN
            self.ret_instruction = reverse_ends(self.ret_instruction)
            self.nop_instruction = reverse_ends(self.nop_instruction)

        # unicorn specific stuff
        if self.uc_mode is not None:
            if endness == 'Iend_BE':
                self.uc_mode -= _unicorn.UC_MODE_LITTLE_ENDIAN
                self.uc_mode += _unicorn.UC_MODE_BIG_ENDIAN
            self.uc_regs = { }
            # map register names to unicorn const
            for r in self.register_names.itervalues():
                reg_name = self.uc_prefix + 'REG_' + r.upper()
                if hasattr(self.uc_const, reg_name):
                    self.uc_regs[r] = getattr(self.uc_const, reg_name)

    def copy(self):
        new_arch = type(self)(self.memory_endness)
        new_arch.vex_archinfo = self.vex_archinfo.copy()

        return new_arch

    def __repr__(self):
        return '<Arch %s (%s)>' % (self.name, self.memory_endness[-2:])

    def __eq__(self, other):
        if not isinstance(other, Arch):
            return False
        return  self.name == other.name and \
                self.bits == other.bits and \
                self.memory_endness == other.memory_endness

    def __ne__(self, other):
        return not self == other

    def __getstate__(self):
        self._cs = None
        return self.__dict__

    def __setstate__(self, data):
        self.__dict__.update(data)

    def gather_info_from_state(self, state):
        info = {}
        for reg in self.persistent_regs:
            info[reg] = state.registers.load(reg)
        return info

    def prepare_state(self, state, info=None):
        if info is not None:
            # TODO: Only do this for PIC!
            for reg in self.persistent_regs:
                if reg in info:
                    state.registers.store(reg, info[reg])

        return state

    def get_default_reg_value(self, register):
        if register == 'sp':
            # Convert it to the corresponding register name
            registers = [r for r, v in self.registers.items() if v[0] == self.sp_offset]
            if len(registers) > 0:
                register = registers[0]
            else:
                return None
        for reg, val, _, _ in self.default_register_values:
            if reg == register:
                return val
        return None

    def struct_fmt(self, size=None):
        fmt = ""
        if size is None:
            size = self.bits

        if self.memory_endness == "Iend_BE":
            fmt += ">"
        else:
            fmt += "<"

        if size == 64:
            fmt += "Q"
        elif size == 32:
            fmt += "I"
        elif size == 16:
            fmt += "H"
        elif size == 8:
            fmt += "B"
        else:
            raise ValueError("Invalid size: Must be a muliple of 8")

        return fmt

    @property
    def bytes(self):
        """
        Return the standard word size in bytes
        """
        return self.bits/8

    # e.g. sizeof['int'] = 4
    sizeof = {}

    @property
    def capstone(self):
        if self.cs_arch is None:
            raise ArchError("Arch %s does not support disassembly with capstone" % self.name)
        if self._cs is None:
            self._cs = _capstone.Cs(self.cs_arch, self.cs_mode)
            self._cs.detail = True
        return self._cs

    @property
    def unicorn(self):
        if _unicorn is None or self.uc_arch is None:
            raise ArchError("Arch %s does not support with unicorn" % self.name)
        # always create a new unicorn instance
        return _unicorn.Uc(self.uc_arch, self.uc_mode)

    def translate_dynamic_tag(self, tag):
        try:
            return self.dynamic_tag_translation[tag]
        except KeyError:
            if isinstance(tag, (int, long)):
                l.error("Please look up and add dynamic tag type %#x for %s", tag, self.name)
            return tag

    def translate_symbol_type(self, tag):
        try:
            return self.symbol_type_translation[tag]
        except KeyError:
            if isinstance(tag, (int, long)):
                l.error("Please look up and add symbol type %#x for %s", tag, self.name)
            return tag

    def translate_register_name(self, offset):
        try:
            return self.register_names[offset]
        except KeyError:
            return str(offset)

    # Determined by watching the output of strace ld-linux.so.2 --list --inhibit-cache
    def library_search_path(self, pedantic=False):
        subfunc = lambda x: x.replace('${TRIPLET}', self.triplet).replace('${ARCH}', self.linux_name)
        path = ['/lib/${TRIPLET}/', '/usr/lib/${TRIPLET}/', '/lib/', '/usr/lib', '/usr/${TRIPLET}/lib/']
        if self.bits == 64:
            path.append('/usr/${TRIPLET}/lib64/')
            path.append('/usr/lib64/')
            path.append('/lib64/')
        elif self.bits == 32:
            path.append('/usr/${TRIPLET}/lib32/')
            path.append('/usr/lib32/')

        if pedantic:
            path = sum([[x + 'tls/${ARCH}/', x + 'tls/', x + '${ARCH}/', x] for x in path], [])
        return map(subfunc, path)

    # various names
    name = None
    vex_arch = None
    qemu_name = None
    ida_processor = None
    linux_name = None
    triplet = None

    # instruction stuff
    max_inst_bytes = None
    ret_instruction = ''
    nop_instruction = ''
    instruction_alignment = None

    # register ofsets
    ip_offset = None
    sp_offset = None
    bp_offset = None
    ret_offset = None

    # memory stuff
    bits = None
    memory_endness = 'Iend_LE'
    register_endness = 'Iend_LE'
    stack_change = None

    # is it safe to cache IRSBs?
    cache_irsb = True

    function_prologs = set()
    function_epilogs = set()

    # Capstone stuff
    cs_arch = None
    cs_mode = None
    _cs = None

    # Unicorn stuff
    uc_arch = None
    uc_mode = None
    uc_const = None
    uc_prefix = None
    uc_regs = None

    call_pushes_ret = False
    initial_sp = 0x7fff0000

    # Difference of the stack pointer after a call instruction (or its equivalent) is executed
    call_sp_fix = 0

    stack_size = 0x8000000

    # Register information
    default_register_values = [ ]
    entry_register_values = { }
    default_symbolic_registers = [ ]
    registers = { }
    register_names = { }
    argument_registers = { }
    persistent_regs = [ ]
    concretize_unique_registers = set() # this is a list of registers that should be concretized, if unique, at the end of each block

    lib_paths = []
    reloc_s_a = []
    reloc_b_a = []
    reloc_s = []
    reloc_copy = []
    reloc_tls_mod_id = []
    reloc_tls_doffset = []
    reloc_tls_offset = []
    dynamic_tag_translation = {}
    symbol_type_translation = {}
    got_section_name = ''

    vex_archinfo = None

def arch_from_id(ident, endness='', bits=''):
    if bits == 64 or (isinstance(bits, str) and '64' in bits):
        bits = 64
    else:
        bits = 32

    endness = endness.lower()
    endness_unsure = False
    if 'lit' in endness:
        endness = 'Iend_LE'
    elif 'big' in endness:
        endness = 'Iend_BE'
    elif 'lsb' in endness:
        endness = 'Iend_LE'
    elif 'msb' in endness:
        endness = 'Iend_BE'
    elif 'le' in endness:
        endness = 'Iend_LE'
    elif 'be' in endness:
        endness = 'Iend_BE'
    elif 'l' in endness:
        endness = 'Iend_LE'
        endness_unsure = True
    elif 'b' in endness:
        endness = 'Iend_BE'
        endness_unsure = True
    else:
        endness = 'Iend_LE'
        endness_unsure = True

    ident = ident.lower()
    if 'ppc64' in ident or 'powerpc64' in ident:
        if endness_unsure:
            endness = 'Iend_BE'
        return ArchPPC64(endness)
    elif 'ppc' in ident or 'powerpc' in ident:
        if endness_unsure:
            endness = 'Iend_BE'
        if bits == 64:
            return ArchPPC64(endness)
        return ArchPPC32(endness)
    elif 'mips' in ident:
        if 'mipsel' in ident:
            if bits == 64:
                return ArchMIPS64('Iend_LE')
            return ArchMIPS32('Iend_LE')
        if endness_unsure:
            if bits == 64:
                return ArchMIPS64('Iend_BE')
            return ArchMIPS32('Iend_BE')
        if bits == 64:
            return ArchMIPS64(endness)
        return ArchMIPS32(endness)
    elif 'arm' in ident or 'thumb' in ident:
        if endness_unsure:
            if 'l' in ident or 'le' in ident:
                endness = 'Iend_LE'
            elif 'b' in ident or 'be' in ident:
                endness = 'Iend_BE'
        if bits == 64:
            return ArchAArch64(endness)
        return ArchARM(endness)
    elif 'aarch' in ident:
        return ArchAArch64(endness)
    elif 'amd64' in ident or ('x86' in ident and '64' in ident) or 'x64' in ident:
        return ArchAMD64('Iend_LE')
    elif '386' in ident or 'x86' in ident or 'metapc' in ident:
        if bits == 64:
            return ArchAMD64('Iend_LE')
        return ArchX86('Iend_LE')

    raise ArchError("Could not parse out arch!")


def reverse_ends(string):
    ise = 'I'*(len(string)/4)
    return _struct.pack('>' + ise, *_struct.unpack('<' + ise, string))

# pylint: disable=unused-import
from .arch_amd64    import ArchAMD64
from .arch_x86      import ArchX86
from .arch_arm      import ArchARM, ArchARMEL, ArchARMHF
from .arch_aarch64  import ArchAArch64
from .arch_ppc32    import ArchPPC32
from .arch_ppc64    import ArchPPC64
from .arch_mips32   import ArchMIPS32
from .arch_mips64   import ArchMIPS64
from .archerror     import ArchError

all_arches = [
    ArchAMD64(), ArchX86(),
    ArchARM('Iend_LE'), ArchARM('Iend_BE'),
    ArchAArch64('Iend_LE'), ArchAArch64('Iend_BE'),
    ArchPPC32('Iend_LE'), ArchPPC32('Iend_BE'),
    ArchPPC64('Iend_LE'), ArchPPC64('Iend_BE'),
    ArchMIPS32('Iend_LE'), ArchMIPS32('Iend_BE'),
    ArchMIPS64('Iend_LE'), ArchMIPS64('Iend_BE')
]



import capstone as _capstone

try:
    import unicorn as _unicorn
except ImportError:
    _unicorn = None

from .arch import Arch

# TODO: determine proper base register (if it exists)
# TODO: handle multiple return registers?
# TODO: which endianness should be default?

class ArchARM(Arch):
    def __init__(self, endness="Iend_LE"):
        super(ArchARM, self).__init__(endness)
        if endness == 'Iend_BE':
            self.function_prologs = {
                r"\xe9\x2d[\x00-\xff][\x00-\xff]",          # stmfd sp!, {xxxxx}
                r"\xe5\x2d\xe0\x04",                        # push {lr}
            }
            self.function_epilogs = {
                r"\xe8\xbd[\x00-\xff]{2}\xe1\x2f\xff\x1e"   # pop {xxx}; bx lr
                r"\xe4\x9d\xe0\x04\xe1\x2f\xff\x1e"         # pop {xxx}; bx lr
            }

    # ArchARM will match with any ARM, but ArchARMEL/ArchARMHF is a mismatch
    def __eq__(self, other):
        # pylint: disable=unidiomatic-typecheck
        if not isinstance(other, ArchARM):
            return False
        if self.memory_endness != other.memory_endness or self.bits != other.bits:
            return False
        if type(self) is type(other):
            return True
        if type(self) is ArchARM or type(other) is ArchARM:
            return True
        return False

    def __getstate__(self):
        self._cs = None
        self._cs_thumb = None
        return self.__dict__

    def __setstate__(self, data):
        self.__dict__.update(data)

    @property
    def capstone(self):
        if self._cs is None:
            self._cs = _capstone.Cs(self.cs_arch, self.cs_mode + _capstone.CS_MODE_ARM)
            self._cs.detail = True
        return self._cs

    @property
    def capstone_thumb(self):
        if self._cs_thumb is None:
            self._cs_thumb = _capstone.Cs(self.cs_arch, self.cs_mode + _capstone.CS_MODE_THUMB)
            self._cs_thumb.detail = True
        return self._cs_thumb

    @property
    def unicorn(self):
        return _unicorn.Uc(self.uc_arch, self.uc_mode + _unicorn.UC_MODE_ARM) if _unicorn is not None else None

    @property
    def unicorn_thumb(self):
        return _unicorn.Uc(self.uc_arch, self.uc_mode + _unicorn.UC_MODE_THUMB) if _unicorn is not None else None

    bits = 32
    vex_arch = "VexArchARM"
    name = "ARMEL"
    qemu_name = 'arm'
    ida_processor = 'armb'
    linux_name = 'arm'
    triplet = 'arm-linux-gnueabihf'
    max_inst_bytes = 4
    ip_offset = 68
    sp_offset = 60
    bp_offset = 60
    ret_offset = 8
    syscall_num_offset = 36
    call_pushes_ret = False
    stack_change = -4
    memory_endness = 'Iend_LE'
    register_endness = 'Iend_LE'
    sizeof = {'int': 32, 'long': 32, 'long long': 64}
    cs_arch = _capstone.CS_ARCH_ARM
    cs_mode = _capstone.CS_MODE_LITTLE_ENDIAN
    _cs_thumb = None
    uc_arch = _unicorn.UC_ARCH_ARM if _unicorn else None
    uc_mode = _unicorn.UC_MODE_LITTLE_ENDIAN if _unicorn else None
    uc_const = _unicorn.arm_const if _unicorn else None
    uc_prefix = "UC_ARM_" if _unicorn else None
    #self.ret_instruction = "\x0E\xF0\xA0\xE1" # this is mov pc, lr
    ret_instruction = "\x1E\xFF\x2F\xE1" # this is bx lr
    nop_instruction = "\x00\x00\x00\x00"
    function_prologs = {
        r"[\x00-\xff][\x00-\xff]\x2d\xe9",          # stmfd sp!, {xxxxx}
        r"\x04\xe0\x2d\xe5",                        # push {lr}
    }
    function_epilogs = {
        r"[\x00-\xff]{2}\xbd\xe8\x1e\xff\x2f\xe1"   # pop {xxx}; bx lr
        r"\x04\xe0\x9d\xe4\x1e\xff\x2f\xe1"         # pop {xxx}; bx lr
    }
    instruction_alignment = 4
    concretize_unique_registers = {64}
    default_register_values = [
        ( 'sp', Arch.initial_sp, True, 'global' ),      # the stack
        ( 'itstate', 0x00000000, False, None )              # part of the thumb conditional flags
    ]
    entry_register_values = {
        'r0': 'ld_destructor'
    }

    default_symbolic_registers = [ 'r0', 'r1', 'r2', 'r3', 'r4', 'r5', 'r6', 'r7', 'r8', 'r9', 'r10', 'r11', 'r12',
                                   'sp', 'lr', 'pc' ]

    register_names = {
        8: 'r0',
        12: 'r1',
        16: 'r2',
        20: 'r3',
        24: 'r4',
        28: 'r5',
        32: 'r6',
        36: 'r7',
        40: 'r8',
        44: 'r9',
        48: 'r10',
        52: 'r11',
        56: 'r12',

        # stack pointer
        60: 'sp',

        # link register
        64: 'lr',

        # program counter
        68: 'pc',

        # condition stuff
        72: 'cc_op',
        76: 'cc_dep1',
        80: 'cc_dep2',
        84: 'cc_ndep',

        88: 'qflag32',
        92: 'geflag0',
        96: 'geflag1',
        100: 'geflag2',
        104: 'geflag3',

        108: 'emnote',
        112: 'cmstart',
        116: 'cmlen',
        120: 'nraddr',
        124: 'ip_at_syscall',

        128: 'd0',
        136: 'd1',
        144: 'd2',
        152: 'd3',
        160: 'd4',
        168: 'd5',
        176: 'd6',
        184: 'd7',
        192: 'd8',
        200: 'd9',
        208: 'd10',
        216: 'd11',
        224: 'd12',
        232: 'd13',
        240: 'd14',
        248: 'd15',
        256: 'd16',
        264: 'd17',
        272: 'd18',
        280: 'd19',
        288: 'd20',
        296: 'd21',
        304: 'd22',
        312: 'd23',
        320: 'd24',
        328: 'd25',
        336: 'd26',
        344: 'd27',
        352: 'd28',
        360: 'd29',
        368: 'd30',
        376: 'd31',

        384: 'fpscr',
        388: 'tpidruro',
        392: 'itstate'
    }

    registers = {
        # GPRs
        'r0': (8, 4),
        'r1': (12, 4),
        'r2': (16, 4),
        'r3': (20, 4),
        'r4': (24, 4),
        'r5': (28, 4),
        'r6': (32, 4),
        'r7': (36, 4),
        'r8': (40, 4),
        'r9': (44, 4),
        'r10': (48, 4),
        'r11': (52, 4),
        'r12': (56, 4),

        # stack pointer
        'sp': (60, 4), 'bp': (60, 4),
        'r13': (60, 4),

        # link register
        'r14': (64, 4),
        'lr': (64, 4),

        # program counter
        'r15': (68, 4),
        'pc': (68, 4),
        'ip': (68, 4),

        # condition stuff
        'cc_op': (72, 4),
        'cc_dep1': (76, 4),
        'cc_dep2': (80, 4),
        'cc_ndep': (84, 4),

        'qflag32': (88, 4),
        'geflag0': (92, 4),
        'geflag1': (96, 4),
        'geflag2': (100, 4),
        'geflag3': (104, 4),

        'emnote': (108, 4),
        'cmstart': (112, 4),
        'cmlen': (116, 4),
        'nraddr': (120, 4),
        'ip_at_syscall': (124, 4),

        'd0': (128, 8),
        'd1': (136, 8),
        'd2': (144, 8),
        'd3': (152, 8),
        'd4': (160, 8),
        'd5': (168, 8),
        'd6': (176, 8),
        'd7': (184, 8),
        'd8': (192, 8),
        'd9': (200, 8),
        'd10': (208, 8),
        'd11': (216, 8),
        'd12': (224, 8),
        'd13': (232, 8),
        'd14': (240, 8),
        'd15': (248, 8),
        'd16': (256, 8),
        'd17': (264, 8),
        'd18': (272, 8),
        'd19': (280, 8),
        'd20': (288, 8),
        'd21': (296, 8),
        'd22': (304, 8),
        'd23': (312, 8),
        'd24': (320, 8),
        'd25': (328, 8),
        'd26': (336, 8),
        'd27': (344, 8),
        'd28': (352, 8),
        'd29': (360, 8),
        'd30': (368, 8),
        'd31': (376, 8),

        'fpscr': (384, 4),
        'tpidruro': (388, 4),
        'itstate': (392, 4)
    }

    argument_registers = {
        registers['r0'][0],
        registers['r1'][0],
        registers['r2'][0],
        registers['r3'][0],
        registers['r4'][0],
        registers['r5'][0],
        registers['r6'][0],
        registers['r7'][0],
        registers['r8'][0],
        registers['r9'][0],
        registers['r10'][0],
        registers['r11'][0],
        registers['r12'][0]
    }

    got_section_name = '.got'
    ld_linux_name = 'ld-linux.so.3'

class ArchARMHF(ArchARM):
    name = 'ARMHF'
    triplet = 'arm-linux-gnueabihf'
    ld_linux_name = 'ld-linux-armhf.so.3'

class ArchARMEL(ArchARM):
    name = 'ARMEL'
    triplet = 'arm-linux-gnueabi'
    ld_linux_name = 'ld-linux.so.3'



import capstone as _capstone

#try:
#   import unicorn as _unicorn
#except ImportError:
#   _unicorn = None

from .arch import Arch

# Note: PowerPC doesn't have pc, so guest_CIA is commented as IP (no arch visible register)
# PowerPC doesn't have stack base pointer, so bp_offset is set to -1 below
# Normally r1 is used as stack pointer

class ArchPPC32(Arch):
    def __init__(self, endness="Iend_LE"):
        super(ArchPPC32, self).__init__(endness)
        if endness == 'Iend_BE':
            self.function_prologs = {
                # stwu r1, -off(r1); mflr r0
                r"\x94\x21[\x00-\xff]{2}\x7c\x08\x02\xa6"
            }
            self.function_epilogs = {
                # mtlr reg; ... ; blr
                r"[\x00-\xff]{2}\x03\xa6([\x00-\xff]{4}){0,6}\x4e\x80\x00\x20"
            }

    bits = 32
    vex_arch = "VexArchPPC32"
    name = "PPC32"
    qemu_name = 'ppc'
    ida_processor = 'ppc'
    linux_name = 'ppc750'   # ?
    triplet = 'powerpc-linux-gnu'
    max_inst_bytes = 4
    ip_offset = 1160
    sp_offset = 20
    bp_offset = 140
    # https://www.ibm.com/developerworks/community/forums/html/topic?id=77777777-0000-0000-0000-000013836863
    # claims that r15 is the base pointer but that is NOT what I see in practice
    ret_offset = 28
    syscall_num_offset = 16
    call_pushes_ret = False
    stack_change = -4
    sizeof = {'int': 32, 'long': 32, 'long long': 64}
    cs_arch = _capstone.CS_ARCH_PPC
    cs_mode = _capstone.CS_MODE_32 + _capstone.CS_MODE_LITTLE_ENDIAN
    # unicorn not supported
    #uc_arch = _unicorn.UC_ARCH_PPC if _unicorn else None
    #uc_mode = (_unicorn.UC_MODE_32 + _unicorn.UC_MODE_LITTLE_ENDIAN) if _unicorn else None
    ret_instruction = "\x20\x00\x80\x4e"
    nop_instruction = "\x00\x00\x00\x60"
    instruction_alignment = 4

    function_prologs = {
        r"[\x00-\xff]{2}\x21\x94\xa6\x02\x08\x7c",     # stwu r1, -off(r1); mflr r0
    }
    function_epilogs = {
        r"\xa6\x03[\x00-\xff]{2}([\x00-\xff]{4}){0,6}\x20\x00\x80\x4e"    # mtlr reg; ... ; blr
    }

    default_register_values = [
        ( 'sp', Arch.initial_sp, True, 'global' ) # the stack
    ]
    entry_register_values = {
        'r3': 'argc',
        'r4': 'argv',
        'r5': 'envp',
        'r6': 'auxv',
        'r7': 'ld_destructor'
    }

    default_symbolic_registers = ['r0', 'r1', 'r2', 'r3', 'r4', 'r5', 'r6', 'r7', 'r8', 'r9', 'r10', 'r11', 'r12',
                                   'r13', 'r14', 'r15', 'r16', 'r17', 'r18', 'r19', 'r20', 'r21', 'r22', 'r23', 'r24',
                                   'r25', 'r26', 'r27', 'r28', 'r29', 'r30', 'r31', 'sp', 'pc']

    register_names = {
        16: 'r0',
        20: 'r1',
        24: 'r2',
        28: 'r3',
        32: 'r4',
        36: 'r5',
        40: 'r6',
        44: 'r7',
        48: 'r8',
        52: 'r9',
        56: 'r10',
        60: 'r11',
        64: 'r12',
        68: 'r13',
        72: 'r14',
        76: 'r15',
        80: 'r16',
        84: 'r17',
        88: 'r18',
        92: 'r19',
        96: 'r20',
        100: 'r21',
        104: 'r22',
        108: 'r23',
        112: 'r24',
        116: 'r25',
        120: 'r26',
        124: 'r27',
        128: 'r28',
        132: 'r29',
        136: 'r30',
        140: 'r31',

        1168: 'pc',
        1172: 'lr'
    }

    registers = {
        'r0': (16, 4),
        'r1': (20, 4), 'sp': (20, 4),
        'r2': (24, 4),
        'r3': (28, 4),
        'r4': (32, 4),
        'r5': (36, 4),
        'r6': (40, 4),
        'r7': (44, 4),
        'r8': (48, 4),
        'r9': (52, 4),
        'r10': (56, 4),
        'r11': (60, 4),
        'r12': (64, 4),
        'r13': (68, 4),
        'r14': (72, 4),
        'r15': (76, 4), 'bp': (76, 4),
        'r16': (80, 4),
        'r17': (84, 4),
        'r18': (88, 4),
        'r19': (92, 4),
        'r20': (96, 4),
        'r21': (100, 4),
        'r22': (104, 4),
        'r23': (108, 4),
        'r24': (112, 4),
        'r25': (116, 4),
        'r26': (120, 4),
        'r27': (124, 4),
        'r28': (128, 4),
        'r29': (132, 4),
        'r30': (136, 4),
        'r31': (140, 4),

        'ip': (1168, 4),
        'pc': (1168, 4),
        'lr': (1172, 4)
    }

    argument_registers = {
        registers['r0'],
        registers['r2'],
        registers['r3'],
        registers['r4'],
        registers['r5'],
        registers['r6'],
        registers['r7'],
        registers['r8'],
        registers['r9'],
        registers['r10'],
        registers['r11'],
        registers['r12'],
        registers['r13'],
        registers['r14'],
        registers['r15'],
        registers['r16'],
        registers['r17'],
        registers['r18'],
        registers['r19'],
        registers['r20'],
        registers['r21'],
        registers['r22'],
        registers['r23'],
        registers['r24'],
        registers['r25'],
        registers['r26'],
        registers['r27'],
        registers['r28'],
        registers['r29'],
        registers['r30'],
        registers['r31'],
    }

    got_section_name = '.plt'
    ld_linux_name = 'ld.so.1'
